#ifndef SETTINGS_HPP_
#define SETTINGS_HPP_
#include "Random.hpp"
#include "OFStream.hpp"
#include <stdio.h>
#include <cassert>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <string>
#include <utility>
#include <string>
#include "LazyFraction.hpp"
#include "Assertion.hpp"

typedef LazyFraction<7,uint16_t,uint16_t> cls_act_t;
typedef LazyFraction<7,uint16_t,uint16_t> var_act_t;

//define exactly one of the following
#define ON_TIE_DECISION_COUNT 	1
//#define ON_TIE_ACTIVITY			1
//#define ON_TIE_BREAK_COUNT 		1


const int EXIT_UNKNOWN 	= 0;
const int EXIT_SAT 		= 10;
const int EXIT_UNSAT 	= 20;

/**
 * Printing Functionality
 *
 */
//#define print_out(format, args...) ( fflush(stdout), fprintf(stderr, format, ## args), fflush(stderr) )

//#define print_file(file, format, args...) ( fprintf(file, format, ## args) )





#ifdef NDEBUG
#define	warrant(exp) exp
#else
#define	warrant(exp) assert(exp)
#endif

//#define DEBUG_FUNC print_out("%s\n",__func__);
#define DEBUG_FUNC (__ASSERT_VOID_CAST (0));


template<class T>
struct Option{
    typedef std::pair<const char*,T>           val_type;
private:
	T* 									       variable;
	const char*                                command_arg;
	int                                        command_len;
	const char*                                description;
	std::vector<val_type >                     possible_vals;

public:

	Option(T* var,T default_val,const char* arg,const char* descr)
		:variable(var),command_arg(arg),command_len(strlen(arg)),description(descr){
	    *variable      = default_val;};

	void addPossibleValue(val_type pval){  possible_vals.push_back(pval);}

	char* canHandle(char* arg){
	      return (strncmp(arg, command_arg, command_len) == 0) ? (arg + command_len) : NULL;}

	void assign(T val){*variable = val;}

	bool findValue(char* arg){
	    if(possible_vals.empty()) return false;
	    for(int i = possible_vals.size();i--;){
	        if(strcmp(possible_vals[i].first,arg) == 0){
	            assign(possible_vals[i].second);
	            return true;
	        }
	    }
	    fprintf(stderr,"c Unknown value %s of option %s \n",arg,command_arg);
	    exit(EXIT_FAILURE);
	    return false;
	}

	void printValue(std::string format){
	    for(int i = possible_vals.size();i--;){
	        if(*variable == possible_vals[i].second){
	        	fprintf(stderr,"\t%-*s \t%s\n",20,command_arg,possible_vals[i].first);
	            return;
	        }
	    }
	    std::string form = "\t%-*s \t" + format + "\n";
	    fprintf(stderr,form.c_str(),20,command_arg,*variable);
	}


	void printDescription(){
		fprintf(stderr,"\t%-*s \t-%s",20,command_arg,description);
	    int vals = possible_vals.size();
	    if(vals){
	    	fprintf(stderr," - possible values are: ");
	        for(int i = 0; i != vals;++i) fprintf(stderr,"| %s",possible_vals[i].first);
	    }
	    fprintf(stderr,"\n");
	}
};



//Polarity Mode
typedef uint PolarityMode;

const PolarityMode Polarity_False       = 0;
const PolarityMode Polarity_True        = 1;
const PolarityMode Polarity_Rand        = 2;
const PolarityMode Polarity_Prev        = 3;
const PolarityMode Polarity_Hook        = 4;
const PolarityMode Polarity_FlpHook     = 5;
//const PolarityMode Polarity_No_Decision = 4;




class Settings{
	bool              print_console;
public:
	bool			  global_make_count;
	bool			  hybrid;
    uint              rand_seed;
    Random            randy;
    bool              print_local;
    uint32_t          timeout;
    uint64_t	  	  max_literals;
    double			  first_phase_timeout_frac;
    //bool              remove_satisfied;
    PolarityMode      polarity_mode;

    const var_act_t   var_decay;       // Inverse of the variable activity decay factor.
    const var_act_t   var_min_inc;
    const var_act_t   var_max_act;

    const cls_act_t   cls_decay;       // Inverse of the clause activity decay factor.
    const cls_act_t   cls_min_inc;
    const cls_act_t   cls_max_act;

    double            random_var_freq;    // The frequency with which the decision heuristic tries to choose a random variable.        (default 0.02)
    int               conflict_max_init;      // The initial restart limit.                                                                (default 100)
    double            conflict_max_inc;        // The factor with which the restart limit is multiplied in each restart.                    (default 1.5)
    double            learntsize_factor;  // The intitial limit for learnt clauses is a factor of the original clauses.                (default 1 / 3)
    double            learntsize_inc;     // The limit for learnt clauses is multiplied with this factor each restart.                 (default 1.1)
    char*             cnf_file_name;
    char*             csv_file_name;
    char*             gnuplot_file_name;
    char*             solver_name;
    std::string       command_line;
    OFStream* 		  console_out;

    //parameter for simplifier
    bool 				simp;
    const bool			simpblocked; //does not print right model in this version
    bool				simpvarelim;
    uint32_t			simptimeout;

    bool 				printmodel;

    std::vector<Option<uint> >   int_options;
    std::vector<Option<bool> >  bool_options;
    std::vector<Option<char*> > string_options;
    std::vector<Option<bool> > simp_options;

    Settings();
    ~Settings();

    bool             read(int argc, char** argv);
    void             printUsage();
    void 			 printCVSVersionNumbers();
};


const int KILO = 1 << 10;
const int MEGA = 1 << 20;


//inline int min_int(const int x, const int y)            { return (y + (-(x < y) & (x - y)));}

//inline int max_int(const int x, const int y)            { return (y + (-(x > y) & (x - y)));}

//inline void set_max_int(uint32_t& to_be_max, const uint32_t cand) { to_be_max += (-(to_be_max < cand) & (cand - to_be_max));}

inline void set_max_int(uint32_t& to_be_max, const uint32_t cand) { if(cand > to_be_max){to_be_max = cand;}}
//inline void swap(int& x, int& y)                        { x ^= y; y ^= x; x ^= y;}

inline uint32_t hash32(const int& x)                    { return 1 << (x & 31);}

inline uint64_t hash64(const int& x)                    { return 1 << (x & 63);}


template<class E, class Pred>
inline bool find(const std::vector<E>& vec,Pred p){

	int32_t i = vec.size();
    while(i--){if(p(vec[i])){ return true;}}

    return false;
}


template<class E, class Pred>
inline void stable_remove_one(std::vector<E>& vec,Pred p){
    assert(!vec.empty());

    int32_t i = vec.size();
    while(i--){if(p(vec[i])){ ++i; break;}}
    assert( i > 0 );

    for(;i < (int32_t)vec.size();++i){vec[i-1]=vec[i];}
    vec.pop_back();
}


template<class E, class P>
inline void unstable_remove_one(std::vector<E>& vec, P p){
    assert(!vec.empty());

    for(int32_t i(vec.size());i--;){
        if(p(vec[i])){
            vec[i] = vec.back(); vec.pop_back();
            return;
    }}
    assert(false);
}

template<class Vec>
inline uint32_t sum_element_sizes(const Vec& vec){
    uint32_t sum = 0;

    for(uint32_t i = 0;i != vec.size(); ++i){
    	sum += vec[i].size();
    }

    return sum;
}

inline char* bname(char* str, char sep = '/'){
    for(char* iter = str; *iter != '\0' ; ){
        if(*iter++ == sep) str = iter;
    }
    return str;
}


#define settings_hpp_version "Settings.hpp:       $Revision: 1.1 $"
const char*  settings_cpp_version();
#endif /*SETTINGS_HPP_*/
