#include "Statistics.hpp"
#include "BasicSolver.hpp"
#include "DMRPSolver.hpp"
#include "ConstraintDB.hpp"
#include "Types.hpp"
#include "Settings.hpp"
#include "IFStream.hpp"
#include "Settings.hpp"
#include "Random.hpp"
//#include "Flag.hpp"
#include "BitVector.hpp"
#include "DecayHeap.hpp"
#include <locale.h>


Statistics::Statistics(Settings& sets):
    settings                        (sets),
    nbr_propagations                ( 0 ),
    nbr_act_decisions               ( 0 ),
    nbr_rnd_decisions               ( 0 ),
    nbr_restarts                    ( 0 ),
    max_backjumping_distance        ( 0 ),
    sum_backjumping                 ( 0 ),
    //nbr_propagate_shortcut_failed   ( 0 ),
    //nbr_propagate_shortcut_worked   ( 0 ),
    //nbr_search_steps_for_watcher    ( 0 ),
    nbr_expensive_cls_bmp           ( 0 ),
    nbr_expensive_var_bmp           ( 0 ),
    nbr_flipped_prop_learnts		(0),
    nbr_learnt_reductions           ( 0 ),
    total_lemma_shortening          ( 0 ),
    nbr_simplify_execs              ( 0 ),
    nbr_removed_cls_lits            ( 0 ),
    nbr_simplify_worked             ( 0 ),
    nbr_bigsimplify_shrinked        ( 0 ),
    max_dec_level                  ( 0 ),
    nbr_prob_cls_decisions          ( 0 ),
    nbr_expensive_cls_stamp			( 0 ),
    nbr_dmrp_calls					(0),
	//new
	nbr_block_clauses				(0),
	nbr_clauses_original			(0),
	nbr_literals_original			(0),
	nbr_clauses_simplified			(0),
	nbr_literals_simplified			(0),
	parse_timestmp                  ( 0.0 ),
	//new
	simplifying_timestmp			( 0.0 ),
    finish_timestmp                 ( 0.0 )
{}


void Statistics::storeBackjumpingDistance(const int dist){
    /*uint32_t index = dist / 50;
    if(index >= backjumping_distances.size()){    backjumping_distances.resize(index+1,0);}
    backjumping_distances[index] += 1;*/

    //nur alternative
    sum_backjumping += dist;
    set_max_int(max_backjumping_distance,dist);
}


double Statistics::cpuSystemTime()
{
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_stime.tv_sec + (double)ru.ru_stime.tv_usec / 1000000.0;
};

double Statistics::cpuUserTime()
{
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000.0;
};

uint64_t Statistics::usedMemoryBytes(){
    char    path[256];
    sprintf(path, "/proc/%d/statm", getpid());
    FILE*   in = fopen(path, "rb");
    if (in == NULL) return 0;
    uint64_t     value;
# if __WORDSIZE == 64
    if(!fscanf(in, "%ld", &value)){value = 0;}
#else
    if(!fscanf(in, "%lld", &value)){value = 0;}
#endif
    fclose(in);
    return value * (uint64_t)getpagesize();
};


void Statistics::printStatsProgressHead(){
	assert(settings.console_out != NULL);
	settings.console_out->writeString("|    Starts   |  Decisions  | Propagations |  Conflicts  |        Literals        |  DB Red");
}


void Statistics::printStatsProgress(const char c, uint64_t literals,uint64_t nbr_conflicts){
	double  used_bytes = (double)Statistics::usedMemoryBytes();

	assert(settings.console_out != NULL);
	settings.console_out->writeString("| %c %4u %6.1fs| %11llu| %12llu| %12llu| %11llu  %6.2f[MB]|%8u",
			c,
            nbr_restarts, cpuUserTime(),
            nbr_rnd_decisions+nbr_act_decisions,
            nbr_propagations,
            nbr_conflicts,
            literals, (used_bytes / (double)MEGA),
            nbr_learnt_reductions);
}



void Statistics::printFinalStatistics(OFStream& out, const BasicSolver* sp){

	assert(sp != NULL);
    finish_timestmp                = cpuUserTime();
    double  cpu_s_time(Statistics::cpuSystemTime());
    double  used_bytes = (double)Statistics::usedMemoryBytes();
    out.writeString("| ");
    out.writeString("|Statistics: ");

	out.writeString("|Variables:                   %u "             ,sp->numberVariables());
	out.writeString("|Total Clauses:               %u "             ,sp->numberAllClauses());
	out.writeString("|Learnt Clauses:              %u "             ,sp->numberLearntLongClses());
	out.writeString("|Binary Clauses:              %u "             ,sp->numberBinClses());
	out.writeString("|CNF Clause Literals:         %llu "           ,sp->numberLiteralsInLongFixClses()+sp->numberLiteralsInShortClses());
	out.writeString("|Learnt Literals:             %llu "           ,sp->numberLiteralsInLongLearntClses());

    if(settings.print_local){  setlocale( LC_ALL, "" );}
    out.writeString("|CPU User Time:               %g [sec]"        ,finish_timestmp);
    out.writeString("|CPU System Time:             %g [sec]"        ,cpu_s_time);

    //simplifier
    out.writeString("|Simplifier Time:             %g [sec]"        ,simplifying_timestmp);

    out.writeString("|Parse Time:                  %g [sec]"        ,parse_timestmp);
    out.writeString("|Used Memory:                 %.2f [MB] "      ,used_bytes / (double)MEGA);

    out.writeString("|Starts:                      %u "           	,nbr_restarts);
    out.writeString("|DMRP Calls:                  %u "             ,nbr_dmrp_calls);
    out.writeString("|Random Decisions:            %llu (%.2f %%)"  ,nbr_rnd_decisions,(100.0*(double)nbr_rnd_decisions)/(double)(nbr_rnd_decisions + nbr_act_decisions));
    out.writeString("|Activity Decisions:          %llu (%.2f %%)"  ,nbr_act_decisions,(100.0*(double)nbr_act_decisions)/(double)(nbr_rnd_decisions + nbr_act_decisions));
    out.writeString("|Conflicts:                   %llu "           ,sp->conflictCount());
    out.writeString("|Propagations:                %llu "           ,nbr_propagations);

    out.writeString("|Expensive Variable Decay:    %u "             ,nbr_expensive_var_bmp);
    out.writeString("|Flipped Props worked:        %u "             ,nbr_flipped_prop_learnts);

    out.writeString("|Expensive Clause Decay:      %u "             ,nbr_expensive_cls_bmp);
    out.writeString("|Expensive Clause Stamp:      %u "             ,nbr_expensive_cls_stamp);
    out.writeString("|Simplify of clause DB:       %u "             ,nbr_simplify_execs);
    out.writeString("|Simplify worked:             %u "             ,nbr_simplify_worked);
    out.writeString("|Big Simplify shrinked:       %u "             ,nbr_bigsimplify_shrinked);
    out.writeString("|Simplify reduced cls lit:    %u "             ,nbr_removed_cls_lits);
    out.writeString("|Reductions of clause DB:     %u "             ,nbr_learnt_reductions);
    out.writeString("|Total lemma shortening:      %llu "           ,total_lemma_shortening);
    out.writeString("|Max Decision Level:          %u "             ,max_dec_level);
    out.writeString("|Problem Cls Decisions:       %u "             ,nbr_prob_cls_decisions);




    //out.writeString("|Propagate Shortcut Failed:   %u "             ,nbr_propagate_shortcut_failed);
    //out.writeString("|Propagate Shortcut Worked:   %u "             ,nbr_propagate_shortcut_worked);
    //out.writeString("|Search Steps Watcher:        %llu "           ,nbr_search_steps_for_watcher);
    out.writeString("|Max Backjumping:             %d"              ,max_backjumping_distance);
    out.writeString("|Sum Backjumping:             %llu"            ,sum_backjumping);

    out.writeString("|Amount of blocking clauses:  %d"              ,nbr_block_clauses);
	out.writeString("|original #Clauses:           %d"              ,nbr_clauses_original);
	out.writeString("|original #Literals:          %d"              ,nbr_literals_original);
	out.writeString("|preprocessed #Clauses:       %d"              ,nbr_clauses_simplified);
	out.writeString("|preprocessed #Literals:      %d"              ,nbr_literals_simplified);
    /*for(uint i = 0; i != backjumping_distances.size();++i){
        if(backjumping_distances[i])
            out.writeString("|                                          %d-%d:%dx, "    ,i*50,(i+1)*50,backjumping_distances[i]);
    }*/
    if(settings.print_local){  setlocale( LC_ALL, "C" );}
}


static inline const char* polarityMode2String(PolarityMode p){
    switch(p){
        case Polarity_False:            return "Polarity False";
        case Polarity_True:             return "Polarity True";
        case Polarity_Rand:             return "Polarity Random";
        case Polarity_Prev:             return "Polarity Previous";
    }
    assert(false);
    return "Polarity Unknown";
};


inline static std::string toString(double d){
	static char buff[15];
	sprintf(buff," %*.3f ",12,d);
	return std::string(buff);
}

inline static std::string toString(uint64_t l){
	static char buff[15];

# if __WORDSIZE == 64
	sprintf(buff," %12lu ",l);
#else
    sprintf(buff," %12llu ",l);
#endif

	return std::string(buff);
}

inline static std::string toString(uint32_t i){
	static char buff[15];
	sprintf(buff," %12u ",i);
	return std::string(buff);
}

inline static std::string toString(int32_t i){
	static char buff[15];
	sprintf(buff," %12d ",i);
	return std::string(buff);
}

void Statistics::printFile(OFStream& out,const char* sep, const BasicSolver* sp, char* result){

	assert(sp != NULL);
    finish_timestmp = Statistics::cpuUserTime();
    double  cpu_s_time(Statistics::cpuSystemTime());
    double  used_bytes = (double)Statistics::usedMemoryBytes();

    out.writeString("#");
    std::string values  = "";

	out.writeString("Variables     %s",sep);  values  += toString(sp->numberVariables()    ) + sep;
	out.writeString("Total Cls     %s",sep);  values  += toString(sp->numberAllClauses() ) 	 + sep;
	out.writeString("Learnt Cls    %s",sep);  values  += toString(sp->numberLearntLongClses()) + sep;
	out.writeString("Binary Cls    %s",sep);  values  += toString(sp->numberBinClses()   )   + sep;
	out.writeString("CNF Literals  %s",sep);  values  += toString(sp->numberLiteralsInLongFixClses()
																+ sp->numberLiteralsInShortClses()) + sep;
	out.writeString("Learnt Lits   %s",sep);  values  += toString(sp->numberLiteralsInLongLearntClses()) + sep;


    out.writeString("User Time     %s",sep);  values  += toString(finish_timestmp            ) + sep;
    out.writeString("System Time   %s",sep);  values  += toString(cpu_s_time                 ) + sep;
    out.writeString("Parse Time    %s",sep);  values  += toString(parse_timestmp             ) + sep;

    //simplifier
    out.writeString("Simplify Time %s",sep);  values  += toString(simplifying_timestmp       ) + sep;

    out.writeString("Memory        %s",sep);  values  += toString(used_bytes / (double)MEGA  ) + sep;

    out.writeString("Starts        %s",sep);  values  += toString(nbr_restarts               ) + sep;
    out.writeString("Random Decs   %s",sep);  values  += toString(nbr_rnd_decisions          ) + sep;
    out.writeString("Activity Decs %s",sep);  values  += toString(nbr_act_decisions          ) + sep;
    out.writeString("Conflicts     %s",sep);  values  += toString(sp->conflictCount()        ) + sep;
    out.writeString("Propagations  %s",sep);  values  += toString(nbr_propagations           ) + sep;

    out.writeString("Exp Var Decay %s",sep);  values  += toString(nbr_expensive_var_bmp      ) + sep;
    out.writeString("Fl Props work %s",sep);  values  += toString(nbr_flipped_prop_learnts   ) + sep;

    out.writeString("Exp Cls Decay %s",sep);  values  += toString(nbr_expensive_cls_bmp      ) + sep;
    out.writeString("Exp Cls Stamp %s",sep);  values  += toString(nbr_expensive_cls_stamp    ) + sep;
    out.writeString("Simp cls db   %s",sep);  values  += toString(nbr_simplify_execs         ) + sep;
    out.writeString("Simp del cls  %s",sep);  values  += toString(nbr_simplify_worked        ) + sep;
    out.writeString("BigSimp Shrin %s",sep);  values  += toString(nbr_bigsimplify_shrinked   ) + sep;
    out.writeString("Simp del lits %s",sep);  values  += toString(nbr_removed_cls_lits       ) + sep;
    out.writeString("Reduce cdb    %s",sep);  values  += toString(nbr_learnt_reductions      ) + sep;
    out.writeString("Lemma shrink  %s",sep);  values  += toString(total_lemma_shortening     ) + sep;

    //out.writeString("Prop wt good  %s",sep);  values  += toString(nbr_propagate_shortcut_failed)+sep;
    //out.writeString("Prop wt bad   %s",sep);  values  += toString(nbr_propagate_shortcut_worked)+sep;
    //out.writeString("Steps Watcher %s",sep);  values  += toString(nbr_search_steps_for_watcher)+ sep;
    out.writeString("Max Dec Level %s",sep);  values  += toString(max_dec_level             ) + sep;
    out.writeString("Pro Cls Dec   %s",sep);  values  += toString(nbr_prob_cls_decisions     ) + sep;
    out.writeString("Max Backjmp   %s",sep);  values  += toString(max_backjumping_distance   ) + sep;
    out.writeString("Sum Backjmp   %s",sep);  values  += toString(sum_backjumping            ) + sep;

    //simplifier:
   out.writeString("#blocking cls %s",sep);  values  += toString(nbr_block_clauses          ) + sep;
   out.writeString("original #cls %s",sep);  values  += toString(nbr_clauses_original       ) + sep;
   out.writeString("original #lit %s",sep);  values  += toString(nbr_literals_original      ) + sep;
   out.writeString("preproc. #cls %s",sep);  values  += toString(nbr_clauses_simplified     ) + sep;
   out.writeString("preproc. #lit %s",sep);  values  += toString(nbr_literals_simplified    ) + sep;

    time_t rawtime; time(&rawtime); struct tm* timeinfo = localtime(&rawtime);
    char* date_time = asctime(timeinfo); *(date_time + strlen(date_time) - 1) = ' ';

    out.writeString("Date          %s",sep);  values  += " \"" + std::string(date_time) + "\" "+ sep;
    out.writeString("Result        %s",sep);  values  += std::string(result                  ) + sep;
    out.writeString("Instance      %s",sep);  values  += " \"" + std::string(settings.cnf_file_name) + "\" " + sep;

    out.writeString(" \n");

    out.writeString("%s \n",values.c_str());

}



void Statistics::printCSV(OFStream& out,const BasicSolver* sp, char* result){

	assert(sp != NULL);
    if(settings.print_local){  setlocale( LC_ALL, "" );}

    printFile(out,"&",sp,result);

    //out.writeString("Max Backjumping & Backjumping Details ... & ");
    /*out.writeString("%d& ",max_backjumping_distance);
    for(uint i = 0; i != backjumping_distances.size();++i){
        if(backjumping_distances[i]) out.writeString("%d-%d:%dx ; " ,i*50,(i+1)*50,backjumping_distances[i]);
    }*/

    if(settings.print_local){  setlocale( LC_ALL, "C" );}
    out.writeString("\n");
}


const char*  statistics_cpp_version(){return "Statistics.cpp:       $Revision: 1.1 $";}
